#include <stddef.h>
#include <assert.h>
#include "libwr2rf/board.h"
#include "libwr2rf/api.h"
#include "regs.h"
#include "wr2rf_rfnco_regs.h"
#include "ipInfo.h"
#include "oc_spi16.h"

uint32_t
libwr2rf_16x32_read32(struct libwr2rf_dev *dev, unsigned off)
{
  uint32_t l, h;

  /* Note: the order is important.  */
  l = libwr2rf_read16 (dev, off);
  h = libwr2rf_read16 (dev, off + 2);
  return (h << 16) | l;
}

void
libwr2rf_16x32_write32(struct libwr2rf_dev *dev, unsigned off, uint32_t val)
{
  /* Note: the order is important.  */
  libwr2rf_write16 (dev, off + 2, val >> 16);
  libwr2rf_write16 (dev, off, val & 0xffff);
}

void
libwr2rf_16x32_write64(struct libwr2rf_dev *dev, unsigned off, uint64_t val)
{
  libwr2rf_16x32_write32(dev, off, val);
  libwr2rf_16x32_write32(dev, off + 4, val >> 32);
}

uint64_t
libwr2rf_be_read64(struct libwr2rf_dev *dev, unsigned off)
{
  uint64_t l, h;

  h = libwr2rf_be_read32 (dev, off);
  l = libwr2rf_be_read32 (dev, off + 4);
  return (h << 32) | l;
}

void
libwr2rf_be_write64(struct libwr2rf_dev *dev, unsigned off, uint64_t val)
{
  libwr2rf_be_write32 (dev, off, val >> 32);
  libwr2rf_be_write32 (dev, off + 4, val & 0xffffffff);
}

uint32_t
libwr2rf_be_read32(struct libwr2rf_dev *dev, unsigned off)
{
  uint32_t l, h;

  h = libwr2rf_read16 (dev, off);
  l = libwr2rf_read16 (dev, off + 2);
  return (h << 16) | l;
}

void
libwr2rf_be_write32(struct libwr2rf_dev *dev, unsigned off, uint32_t val)
{
  libwr2rf_write16 (dev, off, val >> 16);
  libwr2rf_write16 (dev, off + 2, val & 0xffff);
}

uint32_t
libwr2rf_wrc_read32(struct libwr2rf_dev *dev, unsigned off)
{
        return libwr2rf_16x32_read32(dev, WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_WRPC + off);
}

void
libwr2rf_wrc_write32(struct libwr2rf_dev *dev, unsigned off, uint32_t val)
{
        libwr2rf_16x32_write32(dev, WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_WRPC + off, val);
}


void libwr2rf_pll_spi_init(struct libwr2rf_dev *dev)
{
  /* Min period is 25+25ns, so max freq is 20Mhz.  */
  libwr2rf_oc_spi_init(dev, PLL_BASE, 2);
}

uint8_t libwr2rf_pll_read(struct libwr2rf_dev *dev, unsigned reg)
{
  uint32_t rval;
  libwr2rf_oc_spi_txrx16(dev, PLL_BASE, PLL_CS, 16, (reg << 9) | (1 << 8), &rval);
  return rval & 0xff;
}

void libwr2rf_pll_write(struct libwr2rf_dev *dev, unsigned reg, uint8_t value)
{
  libwr2rf_oc_spi_txrx16(dev, PLL_BASE, PLL_CS, 16, (reg << 9) | (value & 0xff),
                         NULL);
}

void libwr2rf_dds_spi_init(struct libwr2rf_dev *dev)
{
  /* Min period is 25+25ns, so max freq is 20Mhz.  */
  libwr2rf_oc_spi_init(dev, DDS_BASE, 2);
}

void libwr2rf_dds_write(struct libwr2rf_dev *dev, unsigned reg, uint64_t value, unsigned nbits)
{
  uint16_t vals[8];

  assert (reg < 32);

  /* Note: valid nbits are: 64, 32, 16.   */
  assert (nbits == 16 || nbits == 32 || nbits == 64);

  vals[0] = value & 0xffff;
  vals[1] = (value >> 16) & 0xffff;
  vals[2] = (value >> 32) & 0xffff;
  vals[3] = (value >> 48) & 0xffff;
  vals[nbits / 16] = reg;

  libwr2rf_oc_spi_txrx128(dev, DDS_BASE, DDS_CS, nbits + 8, vals);
}

uint64_t libwr2rf_dds_read(struct libwr2rf_dev *dev, unsigned reg, unsigned nbits)
{
  uint16_t vals[8];
  uint64_t res;

  assert (reg < 32);

  /* Note: valid nbits are: 64, 32, 16.   */
  assert (nbits == 16 || nbits == 32 || nbits == 64);

  vals[0] = 0;
  vals[1] = 0;
  vals[2] = 0;
  vals[3] = 0;
  vals[nbits / 16] = 0x80 | reg;

  libwr2rf_oc_spi_txrx128(dev, DDS_BASE, DDS_CS, nbits + 8, vals);

  res = (uint64_t)vals[0];
  if (nbits > 16) {
      res |= ((uint64_t)vals[1]) << 16;
  }
  if (nbits > 32) {
    res |= ((uint64_t)vals[2]) << 32;
    res |= ((uint64_t)vals[3]) << 48;
  }

  return res;
}

void libwr2rf_dds_reset(struct libwr2rf_dev *dev, unsigned reset)
{
  unsigned dds_reset_addr = WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_DDS_CTRL;

  if (reset)
    libwr2rf_write16(dev, dds_reset_addr, WR2RF_INIT_REGS_DDS_CTRL_RESET);
  else
    libwr2rf_write16(dev, dds_reset_addr, 0);
}

void libwr2rf_dds_reset_pulse(struct libwr2rf_dev *dev)
{
  libwr2rf_dds_reset(dev, 1);
  libwr2rf_dds_reset(dev, 0);
}

void libwr2rf_dds_ioupdate(struct libwr2rf_dev *dev)
{
  libwr2rf_write16(dev, WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_DDS_CTRL,
                   WR2RF_INIT_REGS_DDS_CTRL_IOUPDATE);
}

unsigned libwr2rf_dds_sync_error(struct libwr2rf_dev *dev)
{
  return libwr2rf_read16(dev, WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_DDS_STATUS)
    & WR2RF_INIT_REGS_DDS_STATUS_SYNC_ERROR;
}

uint8_t libwr2rf_dac_read(struct libwr2rf_dev *dev, unsigned ch, unsigned reg)
{
  uint32_t rval;
  libwr2rf_oc_spi_txrx16(dev, DAC_BASE, ch, 16, 0x8000 | (reg << 8), &rval);
  return rval & 0xff;
}

void libwr2rf_dac_write(struct libwr2rf_dev *dev, unsigned ch, unsigned reg, uint8_t value)
{
  libwr2rf_oc_spi_txrx16(dev, DAC_BASE, ch, 16, (reg << 8) | (value & 0xff),
    NULL);
}

int libwr2rf_dac_iqsetpoint(struct libwr2rf_dev *dev, unsigned ch, unsigned igain, unsigned qgain)
{
    unsigned igain_addr = 0;
    unsigned qgain_addr = 0;

    switch (ch) {
    case 1:
        igain_addr = WR2RF_VME_REGS_CTRL + WR2RF_CTRL_REGS_RF1_IQDAC_IGAIN;
        qgain_addr = WR2RF_VME_REGS_CTRL + WR2RF_CTRL_REGS_RF1_IQDAC_QGAIN;
        break;
    case 2:
        igain_addr = WR2RF_VME_REGS_CTRL + WR2RF_CTRL_REGS_RF2_IQDAC_IGAIN;
        qgain_addr = WR2RF_VME_REGS_CTRL + WR2RF_CTRL_REGS_RF2_IQDAC_QGAIN;
        break;
    default:
      return -1;
    }

    libwr2rf_write16(dev, igain_addr, igain);
    libwr2rf_write16(dev, qgain_addr, qgain);
    return 0;
}

unsigned libwr2rf_read_fwversion(struct libwr2rf_dev *dev)
{
    unsigned addr = WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_HWINFO
        + HWINFO_FIRMWAREVERSION;

    return libwr2rf_be_read32(dev, addr);
}

int libwr2rf_check_fwversion(struct libwr2rf_dev *dev)
{
  return !(libwr2rf_read_fwversion(dev) == 0x00000800);
}

void libwr2rf_read_nco_ipinfo(struct libwr2rf_dev *dev,
                              struct libwr2rf_nco_ipinfo_t* pinfo)
{
  unsigned addr = WR2RF_VME_REGS_CTRL
    + WR2RF_CTRL_REGS_RF1_RFNCO + RFNCO_IPINFO;

  pinfo->std_ver = libwr2rf_16x32_read32(dev, addr + IPINFO_STDVERSION);
  pinfo->ident = libwr2rf_16x32_read32(dev, addr + IPINFO_IDENT);
  pinfo->fw_ver = libwr2rf_16x32_read32(dev, addr + IPINFO_FIRMWAREVERSION);
  pinfo->mem_map_ver = libwr2rf_16x32_read32(dev, addr + IPINFO_MEMMAPVERSION);
}

int libwr2rf_enable_check_ocxo(struct libwr2rf_dev *dev)
{
  /* OCXO warmup time (in sec).  */
  const unsigned warmup_time = 5 * 60;
  unsigned v;

  /* First, enable the ocxo.
     It's a nop if the ocxo was already enabled.  */
  libwr2rf_write16(dev, WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_OCXO_CTRL,
                   WR2RF_INIT_REGS_OCXO_CTRL_ENABLE);

  /* Read ocxo uptime.  */
  v = libwr2rf_read16(dev, WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_OCXO_UPTIME);
  if (v >= warmup_time)
    return 0;
  else
    return warmup_time - v;
}

int libwr2rf_check_link_time(struct libwr2rf_dev *dev)
{
  unsigned status;

  status = libwr2rf_read16
    (dev, WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_WRC_STATUS);
  return (status & WR2RF_INIT_REGS_WRC_STATUS_LINKUP)
    && (status & WR2RF_INIT_REGS_WRC_STATUS_TIME_VALID);
}

int libwr2rf_get_mac(struct libwr2rf_dev *dev, unsigned char *mac)
{
  unsigned h, l;
  libwr2rf_be_write32(dev, WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_WRC_PAGE1,
                      040000);
  h = libwr2rf_wrc_read32(dev, 0x100 + 0x24);
  mac[0] = (h >> 8) & 0xff;
  mac[1] = (h >> 0) & 0xff;
  l = libwr2rf_wrc_read32(dev, 0x100 + 0x28);
  mac[2] = (l >> 24) & 0xff;
  mac[3] = (l >> 16) & 0xff;
  mac[4] = (l >> 8) & 0xff;
  mac[5] = (l >> 0) & 0xff;

  return 0;
}

int libwr2rf_wrc_time_valid(struct libwr2rf_dev *dev)
{
    unsigned status;

    status = libwr2rf_read16(dev, WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_WRC_STATUS);
    return !(status & WR2RF_INIT_REGS_WRC_STATUS_TIME_VALID);
}

int libwr2rf_set_rf_out(struct libwr2rf_dev *dev, unsigned ch, unsigned cfg)
{
  unsigned rf;
  unsigned addr;
  unsigned val;

  switch (ch) {
  case 1:
    rf = WR2RF_INIT_RF_REGS_RF1;
    break;
  case 2:
    rf = WR2RF_INIT_RF_REGS_RF2;
    break;
  default:
    return -1;
  }

  switch (cfg) {
  case LIBWR2RF_RF_OUT_GND:
    val = 0x2;
    break;
  case LIBWR2RF_RF_OUT_MIXER:
    val = WR2RF_INIT_RF_CH_REGS_COMMON_MIXER_EN | 0x6;
    break;
  case LIBWR2RF_RF_OUT_DDS:
    val = 0x4;
    break;
  default:
    return -2;
  }

  addr = WR2RF_VME_REGS_INIT
    + WR2RF_INIT_REGS_RF + rf + WR2RF_INIT_RF_CH_REGS_COMMON;
  libwr2rf_write16(dev, addr, val);

  return 0;
}

static unsigned diag_id_to_addr(unsigned id)
{
  switch (id) {
  case LIBWR2RF_DIAG_ID_RF_1:
    return WR2RF_VME_REGS_CTRL + WR2RF_CTRL_REGS_RF1_VTUS
      + WR2RF_RFTRIGGER_REGS_RF_DIAG;
  case LIBWR2RF_DIAG_ID_RF_2:
    return WR2RF_VME_REGS_CTRL + WR2RF_CTRL_REGS_RF2_VTUS
      + WR2RF_RFTRIGGER_REGS_RF_DIAG;
  case LIBWR2RF_DIAG_ID_RF_1_TRIG_1:
    return WR2RF_VME_REGS_CTRL +  WR2RF_CTRL_REGS_RF1_VTUS
      + WR2RF_RFTRIGGER_REGS_T1 + TRIGUNIT_REGS_TRIGDIAG;
  case LIBWR2RF_DIAG_ID_RF_1_TRIG_2:
    return WR2RF_VME_REGS_CTRL +  WR2RF_CTRL_REGS_RF1_VTUS
      + WR2RF_RFTRIGGER_REGS_T2 + TRIGUNIT_REGS_TRIGDIAG;
  case LIBWR2RF_DIAG_ID_RF_2_TRIG_1:
    return WR2RF_VME_REGS_CTRL +  WR2RF_CTRL_REGS_RF2_VTUS
      + WR2RF_RFTRIGGER_REGS_T1 + TRIGUNIT_REGS_TRIGDIAG;
  case LIBWR2RF_DIAG_ID_RF_2_TRIG_2:
    return WR2RF_VME_REGS_CTRL +  WR2RF_CTRL_REGS_RF2_VTUS
      + WR2RF_RFTRIGGER_REGS_T2 + TRIGUNIT_REGS_TRIGDIAG;
  default:
    return 0;
  }
}

int libwr2rf_enable_diag(struct libwr2rf_dev *dev, unsigned id,
                         unsigned window_us, unsigned *generation)
{
  unsigned addr = diag_id_to_addr(id);
  unsigned window;

  if (addr == 0)
    return -1;

  switch (window_us) {
  case 5:
    window = 0;
    break;
  case 500:
    window = 2;
    break;
  case 50000: /* 50ms */
    window = 4;
    break;
  case 5000000: /* 5sec */
    window = 5;
    break;
  case 20000000: /* 20sec */
    window = 6;
    break;
  default:
    return -2;
  }
  libwr2rf_write16(dev, addr + VTUDIAG_REGS_CONTROL,
                   window | VTUDIAG_REGS_CONTROL_ENABLE);

  *generation = libwr2rf_read16(dev, addr + VTUDIAG_REGS_GENERATION);
  return 0;
}

int libwr2rf_read_diag(struct libwr2rf_dev *dev, unsigned id,
                       unsigned *freq, unsigned *count, unsigned *generation)
{
  unsigned addr = diag_id_to_addr(id);
  unsigned gen, gen1;

  if (addr == 0)
    return -1;

  /* Be sure FREQ and GENERATION match.  */
  gen = libwr2rf_read16(dev, addr + VTUDIAG_REGS_GENERATION);
  while (1) {
    *freq = libwr2rf_be_read32(dev, addr + VTUDIAG_REGS_FREQ);
    gen1 = libwr2rf_read16(dev, addr + VTUDIAG_REGS_GENERATION);
    if (gen == gen1)
      break;
    gen = gen1;
  }

  *count = libwr2rf_be_read32(dev, addr + VTUDIAG_REGS_COUNTER);
  *generation = gen;
  return 0;
}

static unsigned libwr2rf_lemo_to_bit(unsigned lemo)
{
  switch (lemo)
    {
    case LIBWR2RF_LEMO_TMGIO1:
      return 1 << 1;
    case LIBWR2RF_LEMO_TMGIO2:
      return 1 << 2;
    case LIBWR2RF_LEMO_TMGIO3:
      return 1 << 3;
    case LIBWR2RF_LEMO_TMGIO4:
      return 1 << 4;
    case LIBWR2RF_LEMO_TMGCLK1:
      return 1 << 5;
    case LIBWR2RF_LEMO_TMGCLK2:
      return 1 << 6;
    default:
      return 0;
    }
}

int libwr2rf_vtu_start_lemo(struct libwr2rf_dev *dev, unsigned id,
                            unsigned lemo)
{
  unsigned addr = WR2RF_VME_REGS_INIT;
  unsigned val;

  switch (id)
    {
    case LIBWR2RF_TRIG_ID_RF_1_TRIG_1:
      addr += WR2RF_INIT_REGS_LEMO_RF1_T1_START;
      break;
    case LIBWR2RF_TRIG_ID_RF_1_TRIG_2:
      addr += WR2RF_INIT_REGS_LEMO_RF1_T2_START;
      break;
    case LIBWR2RF_TRIG_ID_RF_2_TRIG_1:
      addr += WR2RF_INIT_REGS_LEMO_RF2_T1_START;
      break;
    case LIBWR2RF_TRIG_ID_RF_2_TRIG_2:
      addr += WR2RF_INIT_REGS_LEMO_RF2_T2_START;
      break;
    default:
      return -1;
    }

  val = libwr2rf_lemo_to_bit(lemo);
  if (val == 0)
    return -1;

  libwr2rf_write16(dev, addr, val);

  return 0;
}

int libwr2rf_vtu_stop_lemo(struct libwr2rf_dev *dev, unsigned id,
                            unsigned lemo)
{
  unsigned addr = WR2RF_VME_REGS_INIT;
  unsigned val;

  switch (id)
    {
    case LIBWR2RF_TRIG_ID_RF_1_TRIG_1:
      addr += WR2RF_INIT_REGS_LEMO_RF1_T1_STOP;
      break;
    case LIBWR2RF_TRIG_ID_RF_1_TRIG_2:
      addr += WR2RF_INIT_REGS_LEMO_RF1_T2_STOP;
      break;
    case LIBWR2RF_TRIG_ID_RF_2_TRIG_1:
      addr += WR2RF_INIT_REGS_LEMO_RF2_T1_STOP;
      break;
    case LIBWR2RF_TRIG_ID_RF_2_TRIG_2:
      addr += WR2RF_INIT_REGS_LEMO_RF2_T2_STOP;
      break;
    default:
      return -1;
    }

  val = libwr2rf_lemo_to_bit(lemo);
  if (val == 0)
    return -1;

  libwr2rf_write16(dev, addr, val);

  return 0;
}

static unsigned libwr2rf_vtu_csr(unsigned id)
{
  unsigned csr_addr = WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_RF;

  switch (id) {
  case LIBWR2RF_TRIG_ID_RF_1_TRIG_1:
    return csr_addr + WR2RF_INIT_RF_REGS_RF1 + WR2RF_INIT_RF_CH_REGS_CH;
  case LIBWR2RF_TRIG_ID_RF_1_TRIG_2:
    return csr_addr + WR2RF_INIT_RF_REGS_RF1 + WR2RF_INIT_RF_CH_REGS_CH
      + WR2RF_INIT_RF_CH_REGS_CH_SIZE;
    break;
  case LIBWR2RF_TRIG_ID_RF_2_TRIG_1:
    return csr_addr + WR2RF_INIT_RF_REGS_RF2 + WR2RF_INIT_RF_CH_REGS_CH;
    break;
  case LIBWR2RF_TRIG_ID_RF_2_TRIG_2:
    return csr_addr + WR2RF_INIT_RF_REGS_RF2 + WR2RF_INIT_RF_CH_REGS_CH
      + WR2RF_INIT_RF_CH_REGS_CH_SIZE;
    break;
  default:
    return 0;
  }
}

int libwr2rf_vtu_output_ff_delay(struct libwr2rf_dev *dev, unsigned id,
                                 unsigned val)
{
  unsigned csr_addr;
  unsigned del_addr = WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_RF +
    WR2RF_INIT_RF_REGS_RF_TRIGGER_FF_DELAY;
  unsigned v;

  csr_addr = libwr2rf_vtu_csr(id);
  if (csr_addr == 0)
    return -1;

  v = libwr2rf_read16(dev, csr_addr);
  libwr2rf_write16(dev, csr_addr,
                   v | WR2RF_INIT_RF_CH_REGS_CH_CSR_DELAY_LATCH);
  /* pulse.  */
  libwr2rf_write16(dev, del_addr, val);
  libwr2rf_write16(dev, csr_addr,
                   v & ~WR2RF_INIT_RF_CH_REGS_CH_CSR_DELAY_LATCH);
  libwr2rf_write16(dev, csr_addr,
                   v | WR2RF_INIT_RF_CH_REGS_CH_CSR_DELAY_LATCH);

  return 0;
}

int libwr2rf_vtu_output_pad_delay(struct libwr2rf_dev *dev, unsigned id,
                                  unsigned hc, unsigned odelay)
{
    unsigned baddr;

    switch (id) {
    case LIBWR2RF_TRIG_ID_RF_1_TRIG_1:
        baddr = WR2RF_VME_REGS_CTRL + WR2RF_CTRL_REGS_RF1_VTUS + WR2RF_RFTRIGGER_REGS_T1;
        break;
    case LIBWR2RF_TRIG_ID_RF_1_TRIG_2:
        baddr = WR2RF_VME_REGS_CTRL + WR2RF_CTRL_REGS_RF1_VTUS + WR2RF_RFTRIGGER_REGS_T2;
        break;
    case LIBWR2RF_TRIG_ID_RF_2_TRIG_1:
        baddr = WR2RF_VME_REGS_CTRL + WR2RF_CTRL_REGS_RF2_VTUS + WR2RF_RFTRIGGER_REGS_T1;
        break;
    case LIBWR2RF_TRIG_ID_RF_2_TRIG_2:
        baddr = WR2RF_VME_REGS_CTRL + WR2RF_CTRL_REGS_RF2_VTUS + WR2RF_RFTRIGGER_REGS_T2;
        break;
    default:
        return -1;
    }

    libwr2rf_write16(dev, baddr + TRIGUNIT_REGS_TRIGOHCDELAY, hc);
    libwr2rf_write16(dev, baddr + TRIGUNIT_REGS_TRIGODELAY, odelay);

    return 0;
}

int libwr2rf_vtu_output_enable(struct libwr2rf_dev *dev, unsigned id,
                               unsigned en)
{
  unsigned csr_addr;
  unsigned v;

  csr_addr = libwr2rf_vtu_csr(id);
  if (csr_addr == 0)
    return -1;

  v = libwr2rf_read16(dev, csr_addr);

  if (en)
    v &= ~WR2RF_INIT_RF_CH_REGS_CH_CSR_DELAY_OEN;
  else
    v |= WR2RF_INIT_RF_CH_REGS_CH_CSR_DELAY_OEN;
  libwr2rf_write16(dev, csr_addr, v);

  return 0;
}

int libwr2rf_vtu_output_shaper(struct libwr2rf_dev *dev, unsigned id,
                               unsigned en)
{
  unsigned csr_addr;
  unsigned v;

  csr_addr = libwr2rf_vtu_csr(id);
  if (csr_addr == 0)
    return -1;

  v = libwr2rf_read16(dev, csr_addr);

  if (en)
    v &= ~WR2RF_INIT_RF_CH_REGS_CH_CSR_MUX_SEL;
  else
    v |= WR2RF_INIT_RF_CH_REGS_CH_CSR_MUX_SEL;
  libwr2rf_write16(dev, csr_addr, v);

  return 0;
}

int libwr2rf_vtu_output_reset(struct libwr2rf_dev *dev, unsigned id,
                              unsigned rst)
{
  unsigned csr_addr;
  unsigned v;

  csr_addr = libwr2rf_vtu_csr(id);
  if (csr_addr == 0)
    return -1;

  v = libwr2rf_read16(dev, csr_addr);

  if (rst)
    v |= WR2RF_INIT_RF_CH_REGS_CH_CSR_TRIG_RST;
  else
    v &= ~WR2RF_INIT_RF_CH_REGS_CH_CSR_TRIG_RST;
  libwr2rf_write16(dev, csr_addr, v);

  return 0;
}

static unsigned trig_id_to_addr(unsigned id)
{
  switch (id) {
  case LIBWR2RF_TRIG_ID_RF_1_TRIG_1:
    return WR2RF_VME_REGS_CTRL
      + WR2RF_CTRL_REGS_RF1_VTUS + WR2RF_RFTRIGGER_REGS_T1;
  case LIBWR2RF_TRIG_ID_RF_1_TRIG_2:
    return WR2RF_VME_REGS_CTRL
      + WR2RF_CTRL_REGS_RF1_VTUS + WR2RF_RFTRIGGER_REGS_T2;
  case LIBWR2RF_TRIG_ID_RF_2_TRIG_1:
    return WR2RF_VME_REGS_CTRL
      + WR2RF_CTRL_REGS_RF2_VTUS + WR2RF_RFTRIGGER_REGS_T1;
  case LIBWR2RF_TRIG_ID_RF_2_TRIG_2:
    return WR2RF_VME_REGS_CTRL
      + WR2RF_CTRL_REGS_RF2_VTUS + WR2RF_RFTRIGGER_REGS_T2;
  default:
    return 0;
  }
}

int libwr2rf_vtu_reset (struct libwr2rf_dev *dev, unsigned id, unsigned rst)
{
  unsigned addr = trig_id_to_addr (id);
  unsigned v;

  if (addr == 0)
    return -1;

  v = libwr2rf_read16(dev, addr + TRIGUNIT_REGS_CONTROL);
  if (rst)
    v |= TRIGUNIT_REGS_CONTROL_VTURESET;
  else
    v &= ~TRIGUNIT_REGS_CONTROL_VTURESET;
  libwr2rf_write16(dev, addr + TRIGUNIT_REGS_CONTROL, v);

  return 0;
}

int libwr2rf_vtu_softstop (struct libwr2rf_dev *dev, unsigned id)
{
  unsigned addr = trig_id_to_addr (id);
  unsigned v;

  if (addr == 0)
    return -1;

  v = libwr2rf_read16(dev, addr + TRIGUNIT_REGS_CONTROL);
  v |= TRIGUNIT_REGS_CONTROL_STOPSOFTTRIGGER;
  libwr2rf_write16(dev, addr + TRIGUNIT_REGS_CONTROL, v);

  return 0;
}

int libwr2rf_vtu_softstart (struct libwr2rf_dev *dev, unsigned id)
{
  unsigned addr = trig_id_to_addr (id);
  unsigned v;

  if (addr == 0)
    return -1;

  v = libwr2rf_read16(dev, addr + TRIGUNIT_REGS_CONTROL);
  v |= TRIGUNIT_REGS_CONTROL_STARTSOFTTRIGGER;
  libwr2rf_write16(dev, addr + TRIGUNIT_REGS_CONTROL, v);

  return 0;
}

int libwr2rf_vtu_program_pulse (struct libwr2rf_dev *dev, unsigned id,
                                uint64_t b)
{
  unsigned addr = trig_id_to_addr (id);
  unsigned v;

  if (addr == 0)
    return -1;

  /* Check the vtu can be programmed, ie the offline configuration is
     not valid.  */
  if (libwr2rf_read16(dev, addr + TRIGUNIT_REGS_CONFIGOFFLINE)
      & TRIGUNIT_REGS_CONFIGOFFLINE_VALID)
    return -2;

  libwr2rf_be_write64(dev, addr + TRIGUNIT_REGS_BVALUEOFFLINE, b);

  v = 0x00 << TRIGUNIT_REGS_CONFIGOFFLINE_MODE_SHIFT;
  v |= TRIGUNIT_REGS_CONFIGOFFLINE_VALID;
  libwr2rf_write16(dev, addr + TRIGUNIT_REGS_CONFIGOFFLINE, v);

  return 0;
}

int libwr2rf_vtu_program_infinite (struct libwr2rf_dev *dev, unsigned id,
                                   uint64_t b, unsigned ht)
{
  unsigned addr = trig_id_to_addr (id);
  unsigned v;

  if (addr == 0)
    return -1;

  /* Check the vtu can be programmed, ie the offline configuration is
     not valid.  */
  if (libwr2rf_read16(dev, addr + TRIGUNIT_REGS_CONFIGOFFLINE)
      & TRIGUNIT_REGS_CONFIGOFFLINE_VALID)
    return -2;

  libwr2rf_be_write64(dev, addr + TRIGUNIT_REGS_BVALUEOFFLINE, b);
  libwr2rf_be_write64(dev, addr + TRIGUNIT_REGS_HTVALUEOFFLINE, ht);

  v = 0x01 << TRIGUNIT_REGS_CONFIGOFFLINE_MODE_SHIFT;
  v |= TRIGUNIT_REGS_CONFIGOFFLINE_VALID;
  libwr2rf_write16(dev, addr + TRIGUNIT_REGS_CONFIGOFFLINE, v);

  return 0;
}

int libwr2rf_vtu_program_window (struct libwr2rf_dev *dev, unsigned id,
                                 uint64_t b, unsigned ht, unsigned w)
{
  unsigned addr = trig_id_to_addr (id);
  unsigned v;

  if (addr == 0)
    return -1;

  /* Check the vtu can be programmed, ie the offline configuration is
     not valid.  */
  if (libwr2rf_read16(dev, addr + TRIGUNIT_REGS_CONFIGOFFLINE)
      & TRIGUNIT_REGS_CONFIGOFFLINE_VALID)
    return -2;

  libwr2rf_be_write64(dev, addr + TRIGUNIT_REGS_BVALUEOFFLINE, b);
  libwr2rf_be_write64(dev, addr + TRIGUNIT_REGS_HTVALUEOFFLINE, ht);
  libwr2rf_be_write64(dev, addr + TRIGUNIT_REGS_WVALUEOFFLINE, w);

  v = 0x02 << TRIGUNIT_REGS_CONFIGOFFLINE_MODE_SHIFT;
  v |= TRIGUNIT_REGS_CONFIGOFFLINE_VALID;
  libwr2rf_write16(dev, addr + TRIGUNIT_REGS_CONFIGOFFLINE, v);

  return 0;
}


int libwr2rf_vtu_program_highfreq (struct libwr2rf_dev *dev, unsigned id,
                                   uint64_t b, unsigned div)
{
  unsigned addr = trig_id_to_addr (id);
  unsigned v;

  if (addr == 0)
    return -1;

  /* Check the vtu can be programmed, ie the offline configuration is
     not valid.  */
  if (libwr2rf_read16(dev, addr + TRIGUNIT_REGS_CONFIGOFFLINE)
      & TRIGUNIT_REGS_CONFIGOFFLINE_VALID)
    return -2;

  libwr2rf_be_write64(dev, addr + TRIGUNIT_REGS_BVALUEOFFLINE, b);
  libwr2rf_be_write64(dev, addr + TRIGUNIT_REGS_HTVALUEOFFLINE, div);

  v = 0x06 << TRIGUNIT_REGS_CONFIGOFFLINE_MODE_SHIFT;
  v |= TRIGUNIT_REGS_CONFIGOFFLINE_VALID;
  libwr2rf_write16(dev, addr + TRIGUNIT_REGS_CONFIGOFFLINE, v);

  return 0;
}

int libwr2rf_vtu_program_invalidate (struct libwr2rf_dev *dev, unsigned id)
{
  unsigned addr = trig_id_to_addr (id);
  unsigned v;

  if (addr == 0)
    return -1;

  v = libwr2rf_read16(dev, addr + TRIGUNIT_REGS_CONFIGOFFLINE);
  if (v & TRIGUNIT_REGS_CONFIGOFFLINE_VALID)
    {
      v &= ~TRIGUNIT_REGS_CONFIGOFFLINE_VALID;
      libwr2rf_write16(dev, addr + TRIGUNIT_REGS_CONFIGOFFLINE, v);
    }

  return 0;
}

int libwr2rf_vtu_nco_reset_delay (struct libwr2rf_dev *dev, unsigned id,
                                  unsigned cdelay, unsigned fdelay, unsigned odelay)
{
  unsigned addr;

  addr = WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_RF
    + WR2RF_INIT_RF_CH_REGS_T1_SYNC_PROGDELAY;

  switch (id)
    {
    case LIBWR2RF_RF1_CHANNEL_ID:
      addr = addr + WR2RF_INIT_RF_REGS_RF1;
      break;
    case LIBWR2RF_RF2_CHANNEL_ID:
      addr = addr + WR2RF_INIT_RF_REGS_RF2;
      break;
    default:
      return -1;
    }

  cdelay = (cdelay << WR2RF_INIT_RF_CH_REGS_T1_SYNC_PROGDELAY_CDELAY_SHIFT) &
    WR2RF_INIT_RF_CH_REGS_T1_SYNC_PROGDELAY_CDELAY_MASK;
  fdelay = (fdelay << WR2RF_INIT_RF_CH_REGS_T1_SYNC_PROGDELAY_FDELAY_SHIFT) &
    WR2RF_INIT_RF_CH_REGS_T1_SYNC_PROGDELAY_FDELAY_MASK;
  odelay = (odelay << WR2RF_INIT_RF_CH_REGS_T1_SYNC_PROGDELAY_ODELAY_SHIFT) &
    WR2RF_INIT_RF_CH_REGS_T1_SYNC_PROGDELAY_ODELAY_MASK;

  libwr2rf_write16(dev, addr, odelay | fdelay | cdelay);

  return 0;
}

int libwr2rf_ioupdate_delay (struct libwr2rf_dev *dev,
                                  unsigned fdelay, unsigned odelay)
{
  unsigned addr;

  addr = WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_DDS_IOUPDATE;

  fdelay = (fdelay << WR2RF_INIT_REGS_DDS_IOUPDATE_FDELAY_SHIFT) &
    WR2RF_INIT_REGS_DDS_IOUPDATE_FDELAY_MASK;
  odelay = (odelay << WR2RF_INIT_REGS_DDS_IOUPDATE_ODELAY_SHIFT) &
    WR2RF_INIT_REGS_DDS_IOUPDATE_ODELAY_MASK;

  libwr2rf_write16(dev, addr, odelay | fdelay);

  return 0;
}


int
libwr2rf_configure_tmgio(struct libwr2rf_dev *dev, unsigned lemo, unsigned oe, unsigned term)
{
    unsigned addr = WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_TMG;
    unsigned bval = libwr2rf_read16(dev, addr);
    unsigned mask;

    if (lemo < 1 || lemo > 4)
        return -1;
    // TERM bit
    mask = 1 << (lemo - 1 + WR2RF_INIT_REGS_TMG_IO_TERM_SHIFT);
    if (term)
        bval |= mask;
    else
        bval &= ~mask;

    // OE bit
    mask = 1 << (lemo - 1 + WR2RF_INIT_REGS_TMG_IO_OE_SHIFT);
    if (oe)
        bval |= mask;
    else
        bval &= ~mask;

    libwr2rf_write16(dev, addr, bval);
    return 0;
}

int
libwr2rf_configure_tmgclk(struct libwr2rf_dev *dev, unsigned lemo, unsigned oe, unsigned term)
{
    unsigned addr = WR2RF_VME_REGS_INIT + WR2RF_INIT_REGS_TMG;
    unsigned bval = libwr2rf_read16(dev, addr);
    unsigned mask;

    if (lemo < 1 || lemo > 2)
        return -1;
    // TERM bit
    mask = 1 << (lemo - 1 + WR2RF_INIT_REGS_TMG_CLK_TERM_SHIFT);
    if (term)
        bval |= mask;
    else
        bval &= ~mask;

    // OE bit
    mask = 1 << (lemo - 1 + WR2RF_INIT_REGS_TMG_CLK_OE_SHIFT);
    if (oe)
        bval |= mask;
    else
        bval &= ~mask;

    libwr2rf_write16(dev, addr, bval);
    return 0;
}

int
libwr2rf_configure_lemo_debug(struct libwr2rf_dev *dev, unsigned lemo, unsigned dbg)
{
    unsigned baddr = WR2RF_VME_REGS_INIT;
    unsigned addr  = 0;

    if (dbg < 1 || dbg > 9)
        return -1;

    switch (lemo) {
    case LIBWR2RF_LEMO_TMGIO1:
        addr = WR2RF_INIT_REGS_TMGIO1;
        break;
    case LIBWR2RF_LEMO_TMGIO2:
        addr = WR2RF_INIT_REGS_TMGIO2;
        break;
    case LIBWR2RF_LEMO_TMGIO3:
        addr = WR2RF_INIT_REGS_TMGIO3;
        break;
    case LIBWR2RF_LEMO_TMGIO4:
        addr = WR2RF_INIT_REGS_TMGIO4;
        break;
    case LIBWR2RF_LEMO_TMGCLK1:
        addr = WR2RF_INIT_REGS_TMGCLK1;
        break;
    case LIBWR2RF_LEMO_TMGCLK2:
        addr = WR2RF_INIT_REGS_TMGCLK2;
        break;
    default:
        return -1;
    }

    libwr2rf_write16(dev, baddr + addr, dbg);
    return 0;
}
